# Core packages
import os
from urllib.parse import urlparse

# Third-party packages
import requests
import prometheus_client

# Local packages
from canonicalwebteam.snapstoreapi.exceptions import (
    ApiTimeoutError,
    ApiConnectionError
)


timeout_counter = prometheus_client.Counter(
    'feed_timeouts',
    'A counter of timed out requests',
    ['domain'],
)
connection_failed_counter = prometheus_client.Counter(
    'feed_connection_failures',
    'A counter of requests which failed to connect',
    ['domain'],
)
latency_histogram = prometheus_client.Histogram(
    'feed_latency_seconds',
    'Feed requests retrieved',
    ['domain', 'code'],
    buckets=[0.25, 0.5, 0.75, 1, 2, 3],
)


def get(
        url,
        headers,
        json=None,
        data=None,
        method=None,
        files=None
):
    """
    WARNING: The cache is disabled due to a bug on our caching system.

    Retrieve the response from the requests cache.
    If the cache has expired then it will attempt to update the cache.
    If it gets an error, it will use the cached response, if it exists.
    """

    if method is None:
        method = "POST" if json else "GET"

    # TODO allow user to choose it's own user agent
    storefront_header = 'storefront ({commit_hash};{environment})'.format(
        commit_hash=os.getenv('COMMIT_ID', 'commit_id'),
        environment=os.getenv('ENVIRONMENT', 'devel')
    )
    headers.update({'User-Agent': storefront_header})
    domain = urlparse(url).netloc

    try:
        response = requests.request(
            method=method,
            url=url,
            headers=headers,
            json=json,
            files=files,
            data=data,
            timeout=3
        )
    except requests.exceptions.Timeout:
        timeout_counter.labels(domain=domain).inc()

        raise ApiTimeoutError(
            'The request to {} took longer than 3 seconds'.format(url),
        )
    except requests.exceptions.ConnectionError:
        connection_failed_counter.labels(domain=domain).inc()

        raise ApiConnectionError(
            'Failed to establish connection to {}.'.format(url)
        )

    latency_histogram.labels(domain=domain, code=response.status_code).observe(
        response.elapsed.total_seconds()
    )

    return response



import unittest
import pymacaroons
import responses
import canonicalwebteam.snapstoreapi.public_api as public_api
import canonicalwebteam.snapstoreapi.publisher_api as publisher_api
import canonicalwebteam.snapstoreapi.exceptions as exceptions
import canonicalwebteam.snapstoreapi.metrics.metrics as metrics


# Make sure tests fail on stray responses.
responses.mock.assert_all_requests_are_fired = True


class SimpleTestsApi(unittest.TestCase):

    @responses.activate
    def test_public_api(self):
        url = (
            'https://api.snapcraft.io/api/v1/snaps/search'
            '?confinement=strict,classic&q=&section=featured.'
        )

        payload = {'api': 'works'}
        responses.add(
            responses.GET, url,
            json=payload, status=200)

        toto = public_api.get_searched_snaps('toto', 1, 1)

        self.assertEqual(toto, payload)

    @responses.activate
    def test_publisher_api(self):
        snap_name = "test-snap"

        api_url = 'https://dashboard.snapcraft.io/dev/api/snaps/info/{}'
        api_url = api_url.format(
            snap_name
        )

        payload = {'api': 'yolo'}
        responses.add(
            responses.GET, api_url,
            json=payload, status=200)

        root = pymacaroons.Macaroon('test', 'testing', 'a_key')
        root.add_third_party_caveat('3rd', 'a_caveat-key', 'a_ident')
        discharge = pymacaroons.Macaroon('3rd', 'a_ident', 'a_caveat_key')
        response = publisher_api.get_snap_info('test-snap', {
            'macaroon_root': root.serialize(),
            'macaroon_discharge': discharge.serialize(),
        })

        self.assertEqual(payload, response)

    @responses.activate
    def test_publisher_api_execption(self):
        snap_name = "test-snap"

        api_url = 'https://dashboard.snapcraft.io/dev/api/snaps/info/{}'
        api_url = api_url.format(
            snap_name
        )

        payload = {
            'error_list': [
                {
                    'code': 'user-not-ready',
                    'message': 'missing namespace'
                }
            ]
        }
        responses.add(
            responses.GET, api_url,
            json=payload, status=500)

        root = pymacaroons.Macaroon('test', 'testing', 'a_key')
        root.add_third_party_caveat('3rd', 'a_caveat-key', 'a_ident')
        discharge = pymacaroons.Macaroon('3rd', 'a_ident', 'a_caveat_key')

        with self.assertRaises(exceptions.MissingUsername):
            publisher_api.get_snap_info(
                'test-snap',
                {
                    'macaroon_root': root.serialize(),
                    'macaroon_discharge': discharge.serialize(),
                },
            )


class OsMetricTest(unittest.TestCase):
    def test_build_os_info(self):
        oses = [
            {
                'name': 'test/-',
                'values': ['0.1']
            },
            {
                'name': 'test2/test',
                'values': ['0.5', '0.9']
            }
        ]

        os_metrics = metrics.OsMetric(
            None, oses, None, None)
        expected_result = [
            {
                'name': 'test2 test',
                'value': '0.9'
            },
            {
                'name': 'test',
                'value': '0.1'
            }
        ]

        self.assertEqual(os_metrics.os, expected_result)



class ApiError(Exception):
    """
    Base exception for any errors in the API layer
    """

    pass


class ApiConnectionError(ApiError):
    """
    Communication with the API failed
    """

    pass


class ApiTimeoutError(ApiError):
    """
    Communication with the API timed out
    """

    pass


class MissingUsername(ApiError):
    """
    The user hasn't registed a username
    """

    pass


class AgreementNotSigned(ApiError):
    """
    The user needs to sign the agreement
    """

    pass


class ApiResponseDecodeError(ApiError):
    """
    We failed to properly decode the response from the API
    """

    pass


class ApiResponseError(ApiError):
    """
    The API responded with an error
    """

    def __init__(self, message, status_code):
        self.status_code = status_code
        return super().__init__(message)


class ApiResponseErrorList(ApiResponseError):
    """
    The API responded with a list of errors,
    which are included in self.errors
    """

    def __init__(self, message, status_code, errors):
        self.errors = errors
        return super().__init__(message, status_code)


class MacaroonRefreshRequired(ApiError):
    """
    The macaroon needs to be refreshed
    """
    def __init__(self):
        return super().__init__('The Macaroon needs to be refreshed')



import canonicalwebteam.snapstoreapi.cache as cache
import os

from canonicalwebteam.snapstoreapi import authentication
from canonicalwebteam.snapstoreapi.exceptions import (
    AgreementNotSigned,
    ApiResponseDecodeError,
    ApiResponseError,
    ApiResponseErrorList,
    MacaroonRefreshRequired,
    MissingUsername
)


DASHBOARD_API = os.getenv(
    'DASHBOARD_API',
    'https://dashboard.snapcraft.io/dev/api/',
)

SNAP_PUB_METRICS_URL = ''.join([
    DASHBOARD_API,
    'snaps/metrics',
])
PUB_METRICS_QUERY_HEADERS = {
    'Content-Type': 'application/json'
}

ACCOUNT_URL = ''.join([
    DASHBOARD_API,
    'account',
])

AGREEMENT_URL = ''.join([
    DASHBOARD_API,
    'agreement/'
])

METADATA_QUERY_URL = ''.join([
    DASHBOARD_API,
    'snaps/{snap_id}/metadata',
])

SCREENSHOTS_QUERY_URL = ''.join([
    DASHBOARD_API,
    'snaps/{snap_id}/binary-metadata'
])

SNAP_INFO_URL = ''.join([
    DASHBOARD_API,
    'snaps/info/{snap_name}',
])

REGISTER_NAME_URL = ''.join([
    DASHBOARD_API,
    'register-name/'
])


def process_response(response):
    try:
        body = response.json()
    except ValueError as decode_error:
        api_error_exception = ApiResponseDecodeError(
            'JSON decoding failed: {}'.format(decode_error),
        )
        raise api_error_exception

    if not response.ok:
        if 'error_list' in body:
            for error in body['error_list']:
                if error['code'] == 'user-not-ready':
                    if 'has not signed agreement' in error['message']:
                        raise AgreementNotSigned
                    elif 'missing namespace' in error['message']:
                        raise MissingUsername

            raise ApiResponseErrorList(
                'The api returned a list of errors',
                response.status_code,
                body['error_list']
            )
        else:
            raise ApiResponseError(
                'Unknown error from api',
                response.status_code
            )

    return body


def get_authorization_header(session):
    authorization = authentication.get_authorization_header(
        session['macaroon_root'],
        session['macaroon_discharge']
    )

    return {
        'Authorization': authorization
    }


def get_account(session):
    headers = get_authorization_header(session)

    response = cache.get(
        url=ACCOUNT_URL,
        method='GET',
        headers=headers
    )

    if authentication.is_macaroon_expired(response.headers):
        raise MacaroonRefreshRequired

    return process_response(response)


def get_agreement(session):
    headers = get_authorization_header(session)

    agreement_response = cache.get(
        AGREEMENT_URL,
        headers
    )

    if authentication.is_macaroon_expired(agreement_response.headers):
        raise MacaroonRefreshRequired

    return agreement_response.json()


def post_agreement(session, agreed):
    headers = get_authorization_header(session)

    json = {
        'latest_tos_accepted': agreed
    }
    agreement_response = cache.get(
        AGREEMENT_URL,
        headers,
        json
    )

    if authentication.is_macaroon_expired(agreement_response.headers):
        raise MacaroonRefreshRequired

    return process_response(agreement_response)


def post_username(session, username):
    headers = get_authorization_header(session)
    json = {
        'short_namespace': username
    }
    username_response = cache.get(
        url=ACCOUNT_URL,
        headers=headers,
        json=json,
        method='PATCH'
    )

    if authentication.is_macaroon_expired(username_response.headers):
        raise MacaroonRefreshRequired

    if username_response.status_code == 204:
        return {}
    else:
        return process_response(username_response)


def get_publisher_metrics(session, json):
    authed_metrics_headers = PUB_METRICS_QUERY_HEADERS.copy()
    auth_header = get_authorization_header(session)['Authorization']
    authed_metrics_headers['Authorization'] = auth_header

    metrics_response = cache.get(
        SNAP_PUB_METRICS_URL,
        headers=authed_metrics_headers,
        json=json
    )

    if authentication.is_macaroon_expired(metrics_response.headers):
        raise MacaroonRefreshRequired

    return process_response(metrics_response)


def post_register_name(
        session, snap_name,
        registrant_comment=None, is_private=False, store=None):

    json = {
        'snap_name': snap_name,
    }

    if registrant_comment:
        json['registrant_comment'] = registrant_comment

    if is_private:
        json['is_private'] = is_private

    if store:
        json['store'] = store

    response = cache.get(
        REGISTER_NAME_URL,
        headers=get_authorization_header(session),
        json=json,
        method='POST')

    if authentication.is_macaroon_expired(response.headers):
        raise MacaroonRefreshRequired

    return process_response(response)


def get_snap_info(snap_name, session):
    response = cache.get(
        SNAP_INFO_URL.format(snap_name=snap_name),
        headers=get_authorization_header(session)
    )

    if authentication.is_macaroon_expired(response.headers):
        raise MacaroonRefreshRequired

    return process_response(response)


def get_snap_id(snap_name, session):
    snap_info = get_snap_info(snap_name, session)

    return snap_info['snap_id']


def snap_metadata(snap_id, session, json=None):
    method = "PUT" if json is not None else None

    metadata_response = cache.get(
        METADATA_QUERY_URL.format(snap_id=snap_id),
        headers=get_authorization_header(session),
        json=json,
        method=method
    )

    if authentication.is_macaroon_expired(metadata_response.headers):
        raise MacaroonRefreshRequired

    return process_response(metadata_response)


def snap_screenshots(snap_id, session, data=None, files=None):
    method = None
    files_array = None
    headers = get_authorization_header(session)
    headers['Accept'] = 'application/json'

    if data:
        method = 'PUT'

        files_array = []
        if files:
            for f in files:
                files_array.append(
                    (f.filename, (f.filename, f.stream, f.mimetype))
                )
        else:
            # API requires a multipart request, but we have no files to push
            # https://github.com/requests/requests/issues/1081
            files_array = {'info': ('', data['info'])}
            data = None

    screenshot_response = cache.get(
        SCREENSHOTS_QUERY_URL.format(snap_id=snap_id),
        headers=headers,
        data=data,
        method=method,
        files=files_array
    )

    if authentication.is_macaroon_expired(screenshot_response.headers):
        raise MacaroonRefreshRequired

    return process_response(screenshot_response)



import os
import requests
from pymacaroons import Macaroon
from urllib.parse import urlparse

DASHBOARD_API = os.getenv(
    'DASHBOARD_API',
    "https://dashboard.snapcraft.io/dev/api/",
)
LOGIN_URL = os.getenv(
    'LOGIN_URL',
    "https://login.ubuntu.com",
)


def get_authorization_header(root, discharge):
    """
    Bind root and discharge macaroons and return the authorization header.
    """

    bound = Macaroon.deserialize(root).prepare_for_request(
        Macaroon.deserialize(discharge)
    )

    return 'Macaroon root={}, discharge={}'.format(root, bound.serialize())


def is_authenticated(session):
    """
    Checks if the user is authenticated from the session
    Returns True if the user is authenticated
    """
    return (
        'openid' in session and
        'macaroon_discharge' in session and
        'macaroon_root' in session
    )


def empty_session(session):
    """
    Empty the session, used to logout.
    """
    session.pop('macaroon_root', None)
    session.pop('macaroon_discharge', None)
    session.pop('openid', None)


def get_caveat_id(root):
    """
    Returns the caveat_id generated by the SSO
    """
    location = urlparse(LOGIN_URL).hostname
    caveat, = [
        c for c in Macaroon.deserialize(root).third_party_caveats()
        if c.location == location
    ]

    return caveat.caveat_id


def request_macaroon():
    """
    Request a macaroon from dashboard.
    Returns the macaroon.
    """
    url = ''.join([
        DASHBOARD_API,
        'acl/',
    ])
    response = requests.request(
        url=url,
        method='POST',
        json={
            'permissions': [
                'package_access',
                'package_upload',
                'edit_account'
            ]
        },
        headers={
            'Accept': 'application/json, application/hal+json',
            'Content-Type': 'application/json',
            'Cache-Control': 'no-cache',
        }
    )

    return response.json()['macaroon']


def get_refreshed_discharge(discharge):
    """
    Get a refresh macaroon if the macaroon is not valid anymore.
    Returns the new discharge macaroon.
    """
    url = ''.join([
        LOGIN_URL,
        '/api/v2/tokens/refresh',
    ])
    response = requests.request(
        url=url,
        method='POST',
        json={'discharge_macaroon': discharge},
        headers={
            'Accept': 'application/json, application/hal+json',
            'Content-Type': 'application/json',
            'Cache-Control': 'no-cache',
        }
    )

    return response.json()['discharge_macaroon']


def is_macaroon_expired(headers):
    """
    Returns True if the macaroon needs to be refreshed from
    the header response.
    """
    return headers.get('WWW-Authenticate') == (
            'Macaroon needs_refresh=1')
